@@ -24,12 +24,11 @@
 #endif
 
 #include <stdio.h>
-#include <fcntl.h>
 #include <unistd.h>
 #include <string.h>
 #include <assert.h>
-#include <termios.h>
 #include <ctype.h>
+#include <errno.h>
 
 #include <glib.h>
 
@@ -44,6 +43,7 @@ static void debug_chat(GAtChat *chat, gboolean in, const char *str, gsize len);
 struct at_command {
 	char *cmd;
 	char **prefixes;
+	gboolean expect_pdu;
 	guint id;
 	GAtResultFunc callback;
 	GAtNotifyFunc listing;
@@ -77,7 +77,6 @@ struct _GAtChat {
 	gpointer user_disconnect_data;		/* user disconnect data */
 	struct ring_buffer *buf;		/* Current read buffer */
 	guint read_so_far;			/* Number of bytes processed */
-	gboolean disconnecting;			/* Whether we're disconnecting */
 	GAtDebugFunc debugf;			/* debugging output function */
 	gpointer debug_data;			/* Data to pass to debug func */
 	char *pdu_notify;			/* Unsolicited Resp w/ PDU */
@@ -134,6 +133,7 @@ static gint at_command_compare_by_id(gconstpointer a, gconstpointer b)
 
 static struct at_command *at_command_create(const char *cmd,
 						const char **prefix_list,
+						gboolean expect_pdu,
 						GAtNotifyFunc listing,
 						GAtResultFunc func,
 						gpointer user_data,
@@ -183,6 +183,7 @@ static struct at_command *at_command_create(const char *cmd,
 
 	c->cmd[len+1] = '\0';
 
+	c->expect_pdu = expect_pdu;
 	c->prefixes = prefixes;
 	c->callback = func;
 	c->listing = listing;
@@ -240,20 +241,21 @@ static void g_at_chat_cleanup(GAtChat *chat)
 		chat->wakeup_timer = 0;
 	}
 
+	if (chat->timeout_source) {
+		g_source_remove(chat->timeout_source);
+		chat->timeout_source = 0;
+	}
+
 	g_at_syntax_unref(chat->syntax);
 	chat->syntax = NULL;
+
+	chat->channel = NULL;
 }
 
 static void read_watcher_destroy_notify(GAtChat *chat)
 {
-	chat->read_watch = 0;
-
-	if (chat->disconnecting)
-		return;
-
-	chat->channel = NULL;
-
 	g_at_chat_cleanup(chat);
+	chat->read_watch = 0;
 
 	if (chat->user_disconnect)
 		chat->user_disconnect(chat->user_disconnect_data);
@@ -321,34 +323,36 @@ static void g_at_chat_finish_command(GAtChat *p, gboolean ok,
 						char *final)
 {
 	struct at_command *cmd = g_queue_pop_head(p->command_queue);
+	GSList *response_lines;
 
 	/* Cannot happen, but lets be paranoid */
 	if (!cmd)
 		return;
 
-	if (cmd->callback) {
-		GAtResult result;
-
-		p->response_lines = g_slist_reverse(p->response_lines);
-
-		result.final_or_pdu = final;
-		result.lines = p->response_lines;
-
-		cmd->callback(ok, &result, cmd->user_data);
-	}
-
-	g_slist_foreach(p->response_lines, (GFunc)g_free, NULL);
-	g_slist_free(p->response_lines);
-	p->response_lines = NULL;
-
-	g_free(final);
-
-	at_command_destroy(cmd);
-
 	p->cmd_bytes_written = 0;
 
 	if (g_queue_peek_head(p->command_queue))
 		g_at_chat_wakeup_writer(p);
+
+	response_lines = p->response_lines;
+	p->response_lines = NULL;
+
+	if (cmd->callback) {
+		GAtResult result;
+
+		response_lines = g_slist_reverse(response_lines);
+
+		result.final_or_pdu = final;
+		result.lines = response_lines;
+
+		cmd->callback(ok, &result, cmd->user_data);
+	}
+
+	g_slist_foreach(response_lines, (GFunc)g_free, NULL);
+	g_slist_free(response_lines);
+
+	g_free(final);
+	at_command_destroy(cmd);
 }
 
 struct terminator_info {
@@ -376,6 +380,7 @@ static gboolean g_at_chat_handle_command_response(GAtChat *p,
 {
 	int i;
 	int size = sizeof(terminator_table) / sizeof(struct terminator_info);
+	int hint;
 
 	for (i = 0; i < size; i++) {
 		struct terminator_info *info = &terminator_table[i];
@@ -403,8 +408,18 @@ static gboolean g_at_chat_handle_command_response(GAtChat *p,
 	}
 
 out:
+	if (cmd->listing && cmd->expect_pdu)
+		hint = G_AT_SYNTAX_EXPECT_PDU;
+	else
+		hint = G_AT_SYNTAX_EXPECT_MULTILINE;
+
 	if (p->syntax->set_hint)
-		p->syntax->set_hint(p->syntax, G_AT_SYNTAX_EXPECT_MULTILINE);
+		p->syntax->set_hint(p->syntax, hint);
+
+	if (cmd->listing && cmd->expect_pdu) {
+		p->pdu_notify = line;
+		return TRUE;
+	}
 
 	if (cmd->listing) {
 		GAtResult result;
@@ -460,19 +475,12 @@ done:
 	g_free(str);
 }
 
-static void have_pdu(GAtChat *p, char *pdu)
+static void have_notify_pdu(GAtChat *p, char *pdu, GAtResult *result)
 {
 	GHashTableIter iter;
 	struct at_notify *notify;
 	char *prefix;
 	gpointer key, value;
-	GAtResult result;
-
-	if (!pdu)
-		goto out;
-
-	result.lines = g_slist_prepend(NULL, p->pdu_notify);
-	result.final_or_pdu = pdu;
 
 	g_hash_table_iter_init(&iter, p->notify_list);
 
@@ -486,13 +494,43 @@ static void have_pdu(GAtChat *p, char *pdu)
 		if (!notify->pdu)
 			continue;
 
-		g_slist_foreach(notify->nodes, at_notify_call_callback,
-					&result);
+		g_slist_foreach(notify->nodes, at_notify_call_callback, result);
 	}
+}
+
+static void have_pdu(GAtChat *p, char *pdu)
+{
+	struct at_command *cmd;
+	GAtResult result;
+	gboolean listing_pdu = FALSE;
+
+	if (!pdu)
+		goto err;
+
+	result.lines = g_slist_prepend(NULL, p->pdu_notify);
+	result.final_or_pdu = pdu;
+
+	cmd = g_queue_peek_head(p->command_queue);
+
+	if (cmd && cmd->expect_pdu && p->cmd_bytes_written > 0) {
+		char c = cmd->cmd[p->cmd_bytes_written - 1];
+
+		if (c == '\r')
+			listing_pdu = TRUE;
+	}
+
+	if (listing_pdu) {
+		cmd->listing(&result, cmd->user_data);
+
+		if (p->syntax->set_hint)
+			p->syntax->set_hint(p->syntax,
+						G_AT_SYNTAX_EXPECT_MULTILINE);
+	} else
+		have_notify_pdu(p, pdu, &result);
 
 	g_slist_free(result.lines);
 
-out:
+err:
 	g_free(p->pdu_notify);
 	p->pdu_notify = NULL;
 
@@ -591,7 +629,7 @@ static void new_bytes(GAtChat *p)
 
 	/* We're overflowing the buffer, shutdown the socket */
 	if (ring_buffer_avail(p->buf) == 0)
-		g_at_chat_shutdown(p);
+		g_source_remove(p->read_watch);
 }
 
 static void debug_chat(GAtChat *chat, gboolean in, const char *str, gsize len)
@@ -781,7 +819,7 @@ static gboolean can_write_data(GIOChannel *channel, GIOCondition cond,
 	}
 
 	if (chat->cmd_bytes_written == 0 && wakeup_first == TRUE) {
-		cmd = at_command_create(chat->wakeup, NULL, NULL, NULL,
+		cmd = at_command_create(chat->wakeup, NULL, FALSE, NULL, NULL,
 					NULL, NULL);
 
 		if (!cmd)
@@ -819,7 +857,7 @@ static gboolean can_write_data(GIOChannel *channel, GIOCondition cond,
 			&bytes_written);
 
 	if (err != G_IO_ERROR_NONE) {
-		g_at_chat_shutdown(chat);
+		g_source_remove(chat->read_watch);
 		return FALSE;
 	}
 
@@ -921,42 +959,12 @@ error:
 	return NULL;
 }
 
-static int open_device(const char *device)
+GIOChannel *g_at_chat_get_channel(GAtChat *chat)
 {
-	struct termios ti;
-	int fd;
-
-	fd = open(device, O_RDWR | O_NOCTTY);
-	if (fd < 0)
-		return -1;
-
-	tcflush(fd, TCIOFLUSH);
-
-	/* Switch TTY to raw mode */
-	memset(&ti, 0, sizeof(ti));
-	cfmakeraw(&ti);
-
-	tcsetattr(fd, TCSANOW, &ti);
-
-	return fd;
-}
-
-GAtChat *g_at_chat_new_from_tty(const char *device, GAtSyntax *syntax)
-{
-	GIOChannel *channel;
-	int fd;
-
-	fd = open_device(device);
-	if (fd < 0)
+	if (chat == NULL)
 		return NULL;
 
-	channel = g_io_channel_unix_new(fd);
-	if (!channel) {
-		close(fd);
-		return NULL;
-	}
-
-	return g_at_chat_new(channel, syntax);
+	return chat->channel;
 }
 
 GAtChat *g_at_chat_ref(GAtChat *chat)
@@ -978,12 +986,11 @@ void g_at_chat_unref(GAtChat *chat)
 
 	is_zero = g_atomic_int_dec_and_test(&chat->ref_count);
 
-	if (is_zero) {
-		g_at_chat_shutdown(chat);
+	if (is_zero == FALSE)
+		return;
 
-		g_at_chat_cleanup(chat);
-		g_free(chat);
-	}
+	g_at_chat_shutdown(chat);
+	g_free(chat);
 }
 
 gboolean g_at_chat_shutdown(GAtChat *chat)
@@ -991,12 +998,9 @@ gboolean g_at_chat_shutdown(GAtChat *chat)
 	if (chat->channel == NULL)
 		return FALSE;
 
-	if (chat->timeout_source) {
-		g_source_remove(chat->timeout_source);
-		chat->timeout_source = 0;
-	}
-
-	chat->disconnecting = TRUE;
+	/* Don't trigger user disconnect on shutdown */
+	chat->user_disconnect = NULL;
+	chat->user_disconnect_data = NULL;
 
 	if (chat->read_watch)
 		g_source_remove(chat->read_watch);
@@ -1007,6 +1011,18 @@ gboolean g_at_chat_shutdown(GAtChat *chat)
 	return TRUE;
 }
 
+gboolean g_at_chat_set_syntax(GAtChat *chat, GAtSyntax *syntax)
+{
+	if (chat == NULL)
+		return FALSE;
+
+	g_at_syntax_unref(chat->syntax);
+
+	chat->syntax = g_at_syntax_ref(syntax);
+
+	return TRUE;
+}
+
 gboolean g_at_chat_set_disconnect_function(GAtChat *chat,
 			GAtDisconnectFunc disconnect, gpointer user_data)
 {
@@ -1032,6 +1048,7 @@ gboolean g_at_chat_set_debug(GAtChat *chat, GAtDebugFunc func, gpointer user)
 
 static guint send_common(GAtChat *chat, const char *cmd,
 			const char **prefix_list,
+			gboolean expect_pdu,
 			GAtNotifyFunc listing, GAtResultFunc func,
 			gpointer user_data, GDestroyNotify notify)
 {
@@ -1040,7 +1057,7 @@ static guint send_common(GAtChat *chat, const char *cmd,
 	if (chat == NULL || chat->command_queue == NULL)
 		return 0;
 
-	c = at_command_create(cmd, prefix_list, listing, func,
+	c = at_command_create(cmd, prefix_list, expect_pdu, listing, func,
 				user_data, notify);
 
 	if (!c)
@@ -1060,7 +1077,7 @@ guint g_at_chat_send(GAtChat *chat, const char *cmd,
 			const char **prefix_list, GAtResultFunc func,
 			gpointer user_data, GDestroyNotify notify)
 {
-	return send_common(chat, cmd, prefix_list, NULL, func,
+	return send_common(chat, cmd, prefix_list, FALSE, NULL, func,
 				user_data, notify);
 }
 
@@ -1072,7 +1089,19 @@ guint g_at_chat_send_listing(GAtChat *chat, const char *cmd,
 	if (listing == NULL)
 		return 0;
 
-	return send_common(chat, cmd, prefix_list, listing, func,
+	return send_common(chat, cmd, prefix_list, FALSE, listing, func,
+				user_data, notify);
+}
+
+guint g_at_chat_send_pdu_listing(GAtChat *chat, const char *cmd,
+				const char **prefix_list,
+				GAtNotifyFunc listing, GAtResultFunc func,
+				gpointer user_data, GDestroyNotify notify)
+{
+	if (listing == NULL)
+		return 0;
+
+	return send_common(chat, cmd, prefix_list, TRUE, listing, func,
 				user_data, notify);
 }
 
