@@ -138,7 +138,7 @@ static const char *call_status_to_string(int status)
 	case CALL_STATUS_DIALING:
 		return "dialing";
 	case CALL_STATUS_ALERTING:
-	return "alerting";
+		return "alerting";
 	case CALL_STATUS_INCOMING:
 		return "incoming";
 	case CALL_STATUS_WAITING:
@@ -551,7 +551,7 @@ static gboolean voicecalls_have_active(struct ofono_voicecall *vc)
 	return FALSE;
 }
 
-static gboolean voicecalls_have_connected(struct ofono_voicecall *vc)
+static gboolean voicecalls_can_dtmf(struct ofono_voicecall *vc)
 {
 	GSList *l;
 	struct voicecall *v;
@@ -561,6 +561,10 @@ static gboolean voicecalls_have_connected(struct ofono_voicecall *vc)
 
 		if (v->call->status == CALL_STATUS_ACTIVE)
 			return TRUE;
+
+		/* Connected for 2nd stage dialing */
+		if (v->call->status == CALL_STATUS_ALERTING)
+			return TRUE;
 	}
 
 	return FALSE;
@@ -835,11 +839,31 @@ static DBusMessage *manager_transfer(DBusConnection *conn,
 	return NULL;
 }
 
+static DBusMessage *manager_swap_without_accept(DBusConnection *conn,
+						DBusMessage *msg, void *data)
+{
+	struct ofono_voicecall *vc = data;
+
+	if (vc->flags & VOICECALLS_FLAG_PENDING)
+		return __ofono_error_busy(msg);
+
+	vc->flags |= VOICECALLS_FLAG_PENDING;
+	vc->pending = dbus_message_ref(msg);
+
+	vc->driver->swap_without_accept(vc, generic_callback, vc);
+
+	return NULL;
+}
+
+
 static DBusMessage *manager_swap_calls(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
 	struct ofono_voicecall *vc = data;
 
+	if (vc->driver->swap_without_accept)
+		return manager_swap_without_accept(conn, msg, data);
+
 	if (vc->flags & VOICECALLS_FLAG_PENDING)
 		return __ofono_error_busy(msg);
 
@@ -1076,7 +1100,7 @@ static DBusMessage *manager_tone(DBusConnection *conn,
 		return __ofono_error_not_implemented(msg);
 
 	/* Send DTMFs only if we have at least one connected call */
-	if (!voicecalls_have_connected(vc))
+	if (!voicecalls_can_dtmf(vc))
 		return __ofono_error_failed(msg);
 
 	if (dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &in_tones,
